/* cmParse - parse clone map and create directory scaffold based on it, one
 * directory for each chromosome and one subdirectory for each contig.  Fill
 * in geno.lst, info and map files for each contig. */
#include "common.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "obscure.h"
#include "linefile.h"
#include "portable.h"
#include "errabort.h"
#include "hash.h"

void usage()
/* Describe usage and exit. */
{
errAbort(
   "cmParse - turn clone map into more easily processed\n"
   "form\n"
   "usage:\n"
   "    cmParse input outDir\n");
}

char *finDir = "/projects/compbio/experiments/hg/gs.1/fin/fa";
char *unfinDir = "/projects/compbio/experiments/hg/gs.1/unfin/fa";
char *predraftDir = "/projects/compbio/experiments/hg/gs.1/predraft/fa";
char *missingDir = "/projects/compbio/experiments/hg/gs.1/missing/fa";

FILE *errLog;


struct cloneInfo
/* Information about one clone. */
    {
    struct cloneInfo *next;
    char *acc;
    char *dir;
    int phase;
    int pos;
    bool inMap;
    bool inFfa;
    };

struct hash *cloneHash;	/* A hash with clone accessions for key and cloneInfo's for value. */
struct cloneInfo *cloneList;  /* List of all clones. */

struct cloneInfo *storeCloneInHash(char *acc)
/* Put clone in hash if it isn't already.  Return cloneInfo on clone. */
{
struct hashEl *hel;
struct cloneInfo *ci;

if ((hel = hashLookup(cloneHash, acc)) != NULL)
    return hel->val;
AllocVar(ci);
hel = hashAdd(cloneHash, acc, ci);
ci->acc = hel->name;
slAddHead(&cloneList, ci);
return ci;
}


struct cmChrom
/* Clone map info for one chromosome. */
    {
    struct cmChrom *next;
    char *name;
    struct cmContig *orderedList;
    struct cmContig *randomList;
    };

struct cloneLine
/* A line that refers to a clone. */
    {
    struct cloneLine *next;	/* next in list. */
    struct cloneInfo *clone;    /* Clone this refers to. */
    char *line;                 /* Line in file. */
    };

struct cmContig
/* Clone map info for one contig (possibly with gaps). */
    {
    struct cmContig *next;	/* Next in list. */
    char *name;			/* Name of contig. */
    struct slName *lineList;    /* Lines that describe it in map file. */
    struct cloneLine *cloneList; /* Clones in map file. */
    };

boolean inContig(struct cmContig *contig, struct cloneInfo *clone)
/* Return TRUE if bac already in contig. */
{
struct cloneLine *cl;

for (cl = contig->cloneList; cl != NULL; cl = cl->next)
    {
    if (clone == cl->clone)
	{
	fprintf(errLog, "%s duplicate %s\n", contig->name, clone->acc);
	return TRUE;
	}
    }
return FALSE;
}

void readContigList(struct lineFile *in, struct cmContig **pContigList, bool isCommitChrom)
/* Read in contigs. */
{
boolean gotEnd = FALSE;
char *lastCtgName = "";
char *ctgName;
char *acc;
int lineSize;
char *line;
char *words[16];
int wordCount;
struct cmContig *contig;
struct cloneLine *cl;
struct cloneInfo *ci;
struct slName *cName;
char oline[512];

while (lineFileNext(in, &line, &lineSize))
    {
    if (lineSize >= sizeof(oline))
	errAbort("Line %d of %s too long (%d bytes)", in->lineIx, in->fileName, lineSize);
    memcpy(oline, line, lineSize+1);
    if (startsWith("end human", line))
	{
	gotEnd = TRUE;
	break;
	}
    wordCount = chopLine(oline, words);
    if (words[0][0] == '@' || words[0][0] == '*')
	continue;	
    if (wordCount != 5 && wordCount != 8 && wordCount != 7)
	errAbort("Expecting 5 or 8 words line %d of %s", in->lineIx, in->fileName);
    ctgName = words[3];
    if (!sameString(ctgName, lastCtgName))
	{
	AllocVar(contig);
	slAddTail(pContigList, contig);
	contig->name = cloneString(ctgName);
	lastCtgName = contig->name;
	}
    cName = newSlName(line);
    slAddTail(&contig->lineList, cName);
    acc = words[0];
    if (!sameWord(acc, "COMMITTED") && !isCommitChrom)
	{
	if (wordCount != 8 && wordCount != 7)
	    errAbort("Need 8 words line %d of %s\n", in->lineIx, in->fileName);
	ci = storeCloneInHash(acc);
	ci->inMap = TRUE;
	if (!inContig(contig, ci))
	    {
	    ci->pos = atoi(words[4]);
	    ci->phase = atoi(words[wordCount-1]);
	    AllocVar(cl);
	    cl->clone = ci;
	    cl->line = cName->name;
	    slAddTail(&contig->cloneList, cl);
	    }
	}
    }
if (!gotEnd)
    warn("%s seems to be truncated\n", in->fileName);
}

void readNa(struct lineFile *in, struct cmContig **pContigList)
/* Read "NA" contigs. */
{
boolean gotEnd = FALSE;
char *acc;
int lineSize;
char *line;
char *words[16];
int wordCount;
struct cmContig *contig;
struct cloneLine *cl;
struct cloneInfo *ci;
struct slName *cName;
char oline[512];
int pos = 0;

AllocVar(contig);
slAddTail(pContigList, contig);
contig->name = cloneString("na");
while (lineFileNext(in, &line, &lineSize))
    {
    if (lineSize >= sizeof(oline))
	errAbort("Line %d of %s too long (%d bytes)", in->lineIx, in->fileName, lineSize);
    memcpy(oline, line, lineSize+1);
    if (startsWith("end human", line))
	{
	gotEnd = TRUE;
	break;
	}
    wordCount = chopLine(oline, words);
    if (words[0][0] == '@' || words[0][0] == '*')
	continue;	
    if (wordCount != 5  && wordCount != 4)
	errAbort("Expecting 5 word line %d of %s", in->lineIx, in->fileName);
    cName = newSlName(line);
    slAddTail(&contig->lineList, cName);
    acc = words[0];
    ci = storeCloneInHash(acc);
    ci->inMap = TRUE;
    if (!inContig(contig, ci))
	{
	ci->pos = pos;
	pos += 100;
	ci->phase = atoi(words[wordCount-2]);
	AllocVar(cl);
	cl->clone = ci;
	cl->line = cName->name;
	slAddTail(&contig->cloneList, cl);
	}
    }
if (!gotEnd)
    warn("%s seems to be truncated\n", in->fileName);
}


struct cmChrom *cmParse(char *inName)
/* Parse wash U style clone map into common
 * intermediate format. */
{
struct lineFile *in = lineFileOpen(inName, TRUE);
int lineSize;
char *line;
char *words[16];
int wordCount;
char chromName[32];
char lastChromName[32];
boolean isOrdered;
char *s;
struct cmChrom *chromList = NULL, *chrom;

strcpy(lastChromName, "");
while (lineFileNext(in, &line, &lineSize))
    {
    struct cmContig **pContigList;

    if (!startsWith("start human SUPERLINK", line))
	continue;
    wordCount = chopLine(line, words);
    if (wordCount != 5)
	errAbort("Odd start line %d of %s\n", in->lineIx, in->fileName);
    s = strrchr(words[2], '.');
    if (s == NULL)
	errAbort("Couldn't find chromosome line %d of 5s\n", in->lineIx, in->fileName);
    s += 1;
    strncpy(chromName, s, sizeof(chromName));
    if (!sameString(chromName, lastChromName))
	{
	strcpy(lastChromName, chromName);
	printf("Reading %s\n", chromName);
	AllocVar(chrom);
	chrom->name = cloneString(chromName);
	slAddHead(&chromList, chrom);
	}
    isOrdered = sameWord(words[3], "ORDERED");
    pContigList = (isOrdered ? &chrom->orderedList : &chrom->randomList);
    if (sameString(chromName, "NA"))
	readNa(in, pContigList);
    else
	readContigList(in, pContigList, sameString(chromName, "COMMIT"));
    }
slReverse(&chromList);
return chromList;
}

boolean wannaMakeDir(char *dir)
/* Make dir if possible.  Return FALSE if not. */
{
if (mkdir(dir, 0775) < 0)
    {
    warn("\nCouldn't make directory %s", dir);
    fprintf(errLog, "%s exists\n", dir);
    perror("");
    return FALSE;
    }
return TRUE;
}

void mustMakeDir(char *dir)
/* Make directory or die trying.  Error even if dir exists. */
{
if (!wannaMakeDir(dir))
    {
    noWarnAbort();
    }
}

void mustChangeDir(char *dir)
/* Change directory or die trying. */
{
if (chdir(dir) < 0)
    {
    warn("Couldn't change directory to %s", dir);
    perror("");
    noWarnAbort();
    }
}

void makeCmDir(struct cmChrom *chromList, char *outDir)
/* Make directory full of goodies corresponding to chromList. */
{
struct cmChrom *chrom;
struct cmContig *contig;
struct slName *line;
struct cloneInfo *ci;
struct cloneLine *cl;
int isOrdered;
FILE *info;
FILE *geno;
FILE *map;

mustMakeDir(outDir);
mustChangeDir(outDir);
for (chrom = chromList; chrom != NULL; chrom = chrom->next)
    {
    if (sameString(chrom->name, "COMMIT"))
	continue;
    printf("Writing chromosome %s", chrom->name);
    fflush(stdout);
    mustMakeDir(chrom->name);
    mustChangeDir(chrom->name);
    for (isOrdered = 1; isOrdered >= 0; --isOrdered)
	{
	contig = (isOrdered ? chrom->orderedList : chrom->randomList);
	for ( ; contig != NULL; contig = contig->next)
	    {
	    printf(".");
	    fflush(stdout);
	    if (wannaMakeDir(contig->name))
		{
		mustChangeDir(contig->name);
		info = mustOpen("info", "w");
		geno = mustOpen("geno.lst", "w");
		map = mustOpen("map", "w");
		fprintf(info, "%s %s\n", 
		    contig->name, 
		    (isOrdered ? "ORDERED" : "RANDOM"));

		for (line = contig->lineList; line != NULL; line = line->next)
		    fprintf(map, "%s\n", line->name);
		for (cl = contig->cloneList; cl != NULL; cl = cl->next)
		    {
		    ci = cl->clone;
		    fprintf(info, "%s %d %d\n", ci->acc, ci->pos, ci->phase);
		    if (ci->dir != NULL)
			fprintf(geno, "%s/%s.fa\n", ci->dir, ci->acc);
		    }
		fclose(info);
		fclose(geno);
		fclose(map);
		mustChangeDir("..");
		}
	    }
	}
    mustChangeDir("..");
    printf("\n");
    }
}

void faDirToCloneHash(char *dir)
/* Read in dir and store file names in hash keyed by
 * accession names. */
{
struct slName *list, *el;
struct cloneInfo *ci;

list = listDir(dir, "*.fa");
for (el = list; el != NULL; el = el->next)
    {
    char rootFile[256];
    char *s;
    strcpy(rootFile, el->name);
    s = strrchr(rootFile, '.');
    if (s != NULL) *s = 0;	/* remove suffix. */
    ci = storeCloneInHash(rootFile);
    ci->inFfa = TRUE;
    ci->dir = dir;
    }
}

void listMissing(struct cloneInfo *ciList)
/* List all clones that are in ffa but not map and vice versa. */
{
struct cloneInfo *ci;
int missingFromMapCount = 0;
int missingFromFfaCount = 0;

fprintf(errLog, "Clones missing from map:\n");
for (ci = ciList; ci != NULL; ci = ci->next)
    {
    if (ci->inFfa && !ci->inMap)
	{
	fprintf(errLog, "%s\n", ci->acc);
	++missingFromMapCount;
	}
    }
fprintf(errLog, "Clones missing from ffa's:\n");
for (ci = ciList; ci != NULL; ci = ci->next)
    {
    if ((!ci->inFfa) && ci->inMap)
	{
	fprintf(errLog, "%s\n", ci->acc);
	++missingFromFfaCount;
	}
    }
fprintf(errLog, "%d clones missing from map, %d missing from ffas\n",
	missingFromMapCount, missingFromFfaCount);
fprintf(stdout, "%d clones missing from map, %d missing from ffas\n",
	missingFromMapCount, missingFromFfaCount);
}

void cmProcess(char *inName, char *outDir)
/* Read in file and create corresponding dir full of goodies. */
{
struct cmChrom *chromList;

cloneHash = newHash(0);
if (chromList == NULL)
    errAbort("Nothing to do in %s\n", inName);
chromList = cmParse(inName);
if (missingDir != NULL)
    {
    printf("Reading missing dir...\n");
    faDirToCloneHash(missingDir);
    }
printf("Reading predraft dir...\n");
faDirToCloneHash(predraftDir);
printf("Reading unfin dir...\n");
faDirToCloneHash(unfinDir);
printf("Reading fin dir...\n");
faDirToCloneHash(finDir);
makeCmDir(chromList, outDir);
slReverse(&cloneList);
listMissing(cloneList);
}

int main(int argc, char *argv[])
/* Process command line. */
{
struct cmChrom *chromList = NULL, *chrom;
if (argc != 3)
    usage();
errLog = mustOpen("err.log", "w");
cmProcess(argv[1], argv[2]);
}



