/*****************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
/* Routines for getting variables passed in from web page
 * forms via CGI. */

#include "common.h"
#include "cheapcgi.h"

boolean cgiIsOnWeb()
/* Return TRUE if looks like we're being run as a CGI. */
{
return getenv("REQUEST_METHOD") != NULL;
}

static char *inputString = NULL;

/* For debugging get the unprocessed input. */
char *_cgiRawInput()
{
return inputString;
}

/* Get query string from environment. 
 * To make error handling more flexible 
 * this returns FALSE the first time it's
 * called if no such environment variable.
 * Thereafter it just pretends that the
 * query string exists, but is empty.
 */
static void getQueryInput()
{
inputString = getenv("QUERY_STRING");
if (inputString == NULL)
    errAbort("No QUERY_STRING in environment.");
}

static void getPostInput()
{
long inputSize;
char *s;
long i;
int r;

s = getenv("CONTENT_LENGTH");
if (s == NULL)
    errAbort("No CONTENT_LENGTH in environment.");
if (sscanf(s, "%lu", &inputSize) != 1)
    errAbort("CONTENT_LENGTH isn't a number.");
inputString = needMem((size_t)inputSize+1);
for (i=0; i<inputSize; ++i)
    {
    r = getc(stdin);
    if (r == EOF)
	errAbort("Short POST input.");
    inputString[i] = r;
    }
inputString[inputSize] = 0;
}

static char *cgiInputSource(char *s)
/* For NULL sources make a guess as to real source. */
{
char *qs;
if (s != NULL)
    return s;
qs = getenv("QUERY_STRING");
if (qs == NULL || qs[0] == 0)
    return "POST";
return "QUERY";
}

void _cgiFindInput(char *method)
{
if (inputString == NULL)
    {
    method = cgiInputSource(method);
    if (sameWord(method, "POST"))
        getPostInput();
    else if (sameWord(method, "QUERY") || sameWord(method, "GET"))
        getQueryInput();
    else
        errAbort("Unknown form method");
    }
}

/* Get the string associated with varName from the 
 * query string. */
static boolean findVarData(char *varName, char **pData, int *pDataLength)
{
char *namePt, *dataPt, *nextNamePt;
int nameLen, dataLen;
int varNameLen = strlen(varName);

if (inputString == NULL)
    _cgiFindInput(NULL);
namePt = inputString;
for (;;)
    {
    dataPt = strchr(namePt, '=');
    if (dataPt == NULL)
	return FALSE;
    nameLen = dataPt - namePt;
    dataPt += 1;
    nextNamePt = strchr(dataPt, '&');
    if (nameLen == varNameLen && strncmp(namePt, varName, nameLen) == 0)
	{
	if (nextNamePt == NULL)
	    dataLen = strlen(dataPt);
	else
	    dataLen = nextNamePt - dataPt;
	*pData = dataPt;
	*pDataLength = dataLen;
	return TRUE;
	}
    if (nextNamePt == NULL)
	return FALSE;
    namePt = nextNamePt + 1;
    }
}

void cgiBadVar(char *varName)
/* Complain about a variable that's not there. */
{
if (varName == NULL) varName = "";
errAbort("Sorry, didn't find input variable %s\n"
        "Probably the web page didn't mean to call this program.", 
        varName);
}

static void mustFindVarData(char *varName, char **pData, int *pDataLength)
{
if (!findVarData(varName, pData, pDataLength))
    cgiBadVar(varName);
}

void cgiDecode(char *in, char *out, int inLength)
/* Decode from cgi pluses-for-spaces format to normal. 
 * Out will be a little shorter than in typically. */
{
char c;
int i;
for (i=0; i<inLength;++i)
    {
    c = *in++;
    if (c == '+') 
	*out++ = ' ';
    else if (c == '%')
	{
	int code;
	if (sscanf(in, "%2x", &code) != 1) 
	    code = '?';
	in += 2;
	i += 2;
	*out++ = code;
	}
    else
	*out++ = c;
    }
*out++ = 0;
}

char *cgiEncode(char *inString)
/* Return a cgi-encoded version of inString. 
 * Alphanumerics kept as is, space translated to plus,
 * and all other characters translated to %hexVal. */
{
char c;
int outSize = 0;
char *outString, *out, *in;

/* Count up how long it will be */
in = inString;
while ((c = *in++) != 0)
    {
    if (isalnum(c) || c == ' ' || c == '.')
        outSize += 1;
    else
        outSize += 3;
    }
outString = needMem(outSize+1);

/* Encode string */
in = inString;
out = outString;
while ((c = *in++) != 0)
    {
    if (isalnum(c) || c == '.')
        *out++ = c;
    else if (c == ' ')
        *out++ = '+';
    else
        {
        unsigned char uc = c;
        char buf[4];
        *out++ = '%';
        sprintf(buf, "%02X", uc);
        *out++ = buf[0];
        *out++ = buf[1];
        }
    }
*out++ = 0;
return outString;
}

char *cgiOptionalString(char *varName)
/* Return value of string if it exists in cgi environment, else NULL */
{
char *data;
char *value;
int dataLength;

if (!findVarData(varName, &data, &dataLength))
    return NULL;
value = needMem(dataLength+1);
cgiDecode(data,value,dataLength);
return value;
}


char *cgiString(char *varName)
{
char *s = cgiOptionalString(varName);
if (s == NULL)
    {
    cgiBadVar(varName);
    }
return s;
}

int cgiInt(char *varName)
{
char *data;
int dataLength;
int x;

mustFindVarData(varName, &data, &dataLength);
if (sscanf(data, "%d", &x)<1) 
     errAbort("Expecting number in %s, got \"%s\"\n", varName, data);
return x;
}

int cgiOptionalInt(char *varName, int defaultVal)
/* This returns integer value of varName if it exists in cgi environment
 * and it's not just the empty string otherwise it returns defaultVal. */
{
char *s;
if ((s = cgiOptionalString(varName)) == NULL)
    return defaultVal;
s = skipLeadingSpaces(s);
if (s[0] == 0)
    return defaultVal;
if (!isdigit(s[0]) && !s[0] == '-')
     errAbort("Expecting number in %s, got \"%s\"\n", varName, s);
return atoi(s);
}

double cgiDouble(char *varName)
{
char *data;
int dataLength;
double x;

mustFindVarData(varName, &data, &dataLength);
if (sscanf(data, "%lf", &x)<1) 
     errAbort("Expecting real number in %s, got \"%s\"\n", varName, data);;
return x;
}

boolean cgiVarExists(char *varName)
/* Returns TRUE if the variable was passed in. */
{
char *data;
int dataLength;

return findVarData(varName, &data, &dataLength);
}

boolean cgiBoolean(char *varName)
{
return cgiVarExists(varName);
}

int cgiOneChoice(char *varName, struct cgiChoice *choices, int choiceSize)
/* Returns value associated with string variable in choice table. */
{
char *key = cgiString(varName);
int i;
int val = -1;

for (i=0; i<choiceSize; ++i)
    {
    if (!differentWord(choices[i].name, key))
        {
        val =  choices[i].value;
        freeMem(key);
        return val;
        }
    }
errAbort("Unknown key %s for variable %s\n", key, varName);
freeMem(key);
return val;
}

void cgiMakeHiddenVar(char *varName, char *string)
/* Store string in hidden input for next time around. */
{
printf("<INPUT TYPE=HIDDEN NAME=\"%s\" VALUE=\"%s\">", varName, string);
}

void cgiContinueHiddenVar(char *varName)
/* Write CGI var back to hidden input for next time around. */
{
if (cgiVarExists(varName))
    cgiMakeHiddenVar(varName, cgiString(varName));
}

boolean cgiSpoof(int *pArgc, char *argv[])
/* Use the command line to set up things as if we were a CGI program. 
 * User types in command line (assuming your program called cgiScript) 
 * like:
 *        cgiScript nonCgiArg1 var1=value1 var2=value2 var3=value3 nonCgiArg2
 * or like
 *        cgiScript nonCgiArg1 var1=value1&var2=value2&var3=value3 nonCgiArg2
 * (The non-cgi arguments can occur anywhere.  The cgi arguments (all containing
 * the character '=') are erased from argc/argv.  Normally you call this
 *        cgiSpoof(&argc, argv);
 */
{
int argc = *pArgc;
int i;
int argcLeft = argc;
char *name;
static char queryString[2046];
char *q = queryString;
boolean needAnd = FALSE;
boolean gotAny = FALSE;

q += sprintf(q, "%s", "QUERY_STRING=");
for (i=0; i<argcLeft; )
    {
    name = argv[i];
    if (strchr(name,'=') != NULL)
        {
        if (needAnd)
            *q++ = '&';
        q += sprintf(q, "%s", name);
        if (strchr(name, '&') == NULL)
            needAnd = TRUE;
        memcpy(&argv[i], &argv[i+1], sizeof(argv[i]) * (argcLeft-i-1));
        argcLeft -= 1;
        gotAny = TRUE;
        }
    else
        i++;
    }
if (gotAny)
    {
    *pArgc = argcLeft;
    putenv("REQUEST_METHOD=GET");
    putenv(queryString);
    }
return gotAny;
}

